/*****************************************************************************
 * vim:tw=78
 *
 * Copyright (C) 2021 Bjarne von Horn (vh at igh dot de).
 *
 * This file is part of the PdCom library.
 *
 * The PdCom library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * The PdCom library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the PdCom library. If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <algorithm>
#include <cstring>
#include <exception>
#include <gnutls/gnutls.h>
#include <pdcom5/Exception.h>
#include <pdcom5/SecureProcess.h>

using PdCom::SecureProcess;

namespace {
template <typename T, typename... Args>
auto wrap(const char *msg, T func, Args &&...args)
        -> decltype(func(std::forward<Args>(args)...))
{
    const auto ans = func(std::forward<Args>(args)...);
    if (ans < 0 && gnutls_error_is_fatal(static_cast<int>(ans)))
        throw PdCom::TlsError(msg, static_cast<int>(ans));
    return ans;
}
}  // namespace

struct PDCOM5_LOCAL SecureProcess::Impl
{
    struct PDCOM5_LOCAL TlsDeleter
    {
        void operator()(gnutls_session_t s)
        {
            if (s)
                gnutls_deinit(s);
        }
        void operator()(gnutls_certificate_credentials_t c)
        {
            if (c)
                gnutls_certificate_free_credentials(c);
        }
    };
    Impl(SecureProcess::EncryptionDetails const &ed, SecureProcess *This);
    std::string host_;
    std::unique_ptr<gnutls_certificate_credentials_st, TlsDeleter> cred_;
    std::unique_ptr<gnutls_session_int, TlsDeleter> session_;
    std::exception_ptr ex_ptr_;
};

SecureProcess::SecureProcess(SecureProcess::EncryptionDetails const &ed) :
    secure_impl_(std::make_shared<Impl>(ed, this))
{}


SecureProcess::SecureProcess(SecureProcess &&o) noexcept :
    secure_impl_(std::move(o.secure_impl_))
{
    if (secure_impl_)
        gnutls_session_set_ptr(secure_impl_->session_.get(), this);
}

SecureProcess &SecureProcess::operator=(SecureProcess &&o) noexcept
{
    if (&o == this)
        return *this;
    std::swap(o.secure_impl_, secure_impl_);
    if (o.secure_impl_)
        gnutls_session_set_ptr(o.secure_impl_->session_.get(), &o);
    if (secure_impl_)
        gnutls_session_set_ptr(secure_impl_->session_.get(), this);
    return *this;
}

int SecureProcess::read(char *const buf, int const count)
{
    const ssize_t recv_chars =
            gnutls_record_recv(secure_impl_->session_.get(), buf, count);
    if (secure_impl_->ex_ptr_)
        std::rethrow_exception(secure_impl_->ex_ptr_);
    if (recv_chars >= 0)
        return static_cast<int>(recv_chars);

    if (recv_chars == GNUTLS_E_REHANDSHAKE) {
        handshake();
        return -EAGAIN;
    }
    if (!gnutls_error_is_fatal(static_cast<int>(recv_chars))) {
        return -EAGAIN;
    }
    throw TlsError("gnutls_record_recv() failed", recv_chars);
}

void SecureProcess::write(const char *buf, size_t count)
{
    if (count == 0)
        return;
    const auto ans =
            gnutls_record_send(secure_impl_->session_.get(), buf, count);
    if (secure_impl_->ex_ptr_)
        std::rethrow_exception(secure_impl_->ex_ptr_);
    if (ans < 0)
        throw TlsError(
                "SecureProcess::secureWrite() failed", static_cast<int>(ans));
    if (static_cast<size_t>(ans) != count)
        throw TlsError("short write in SecureProcess::secureWrite", 0);
}

SecureProcess::Impl::Impl(
        SecureProcess::EncryptionDetails const &ed,
        SecureProcess *This) :
    host_(ed.server_hostname_)
{
    {
        gnutls_session_t s = nullptr;
        wrap("gnutls_init() failed", gnutls_init, &s,
             GNUTLS_CLIENT | GNUTLS_NONBLOCK);
        session_.reset(s);
    }

    wrap("gnutls_priority_set_direct failed", gnutls_priority_set_direct,
         session_.get(), "NORMAL", nullptr);


    gnutls_certificate_credentials_t cc = nullptr;
    wrap("gnutls_certificate_allocate_credentials() failed",
         gnutls_certificate_allocate_credentials, &cc);
    cred_.reset(cc);
    const auto to_datum_t = [](const std::string &s) -> gnutls_datum_t {
        return {const_cast<unsigned char *>(
                        reinterpret_cast<const unsigned char *>(s.data())),
                static_cast<unsigned>(s.size())};
    };
    if (!ed.server_ca_.empty()) {
        const gnutls_datum_t f = to_datum_t(ed.server_ca_);
        const auto a =
                wrap("reading CA failed", gnutls_certificate_set_x509_trust_mem,
                     cc, &f, GNUTLS_X509_FMT_PEM);
        if (a < 1)
            throw TlsError("Less than one CA loaded", 0);
    }
    if (!ed.client_key_.empty() and !ed.client_cert_.empty()) {
        const auto key  = to_datum_t(ed.client_key_);
        const auto cert = to_datum_t(ed.client_cert_);
        wrap("importing client certificate failed",
             gnutls_certificate_set_x509_key_mem, cc, &cert, &key,
             GNUTLS_X509_FMT_PEM);
    }
    wrap("using Certificates failed", gnutls_credentials_set, session_.get(),
         GNUTLS_CRD_CERTIFICATE, cred_.get());

    gnutls_transport_set_push_function(
            session_.get(),
            [](gnutls_transport_ptr_t uptr, const void *data,
               size_t count) -> ssize_t {
                SecureProcess &sp = *reinterpret_cast<SecureProcess *>(uptr);
                try {
                    sp.secureWrite(reinterpret_cast<const char *>(data), count);
                }
                catch (std::exception const &e) {
                    sp.secure_impl_->ex_ptr_ = std::current_exception();
                    // make sure that errno != EAGAIN;
                    errno = EIO;
                    return -1;
                }
                return count;
            });
    gnutls_transport_set_pull_function(
            session_.get(),
            [](gnutls_transport_ptr_t uptr, void *data,
               size_t count) -> ssize_t {
                SecureProcess &sp = *reinterpret_cast<SecureProcess *>(uptr);
                try {
                    const int ans = sp.secureRead(
                            reinterpret_cast<char *>(data), count);
                    if (ans >= 0)
                        return ans;
                    if (ans == -EAGAIN)
                        errno = EAGAIN;
                    return -1;
                }
                catch (std::exception const &e) {
                    sp.secure_impl_->ex_ptr_ = std::current_exception();
                    // make sure that errno != EAGAIN;
                    errno = EIO;
                    return -1;
                }
            });
    if (!host_.empty())
        gnutls_session_set_verify_cert(session_.get(), host_.c_str(), 0);

    gnutls_transport_set_ptr(session_.get(), This);
}

bool SecureProcess::handshake()
{
    const auto a = gnutls_handshake(secure_impl_->session_.get());
    if (a == GNUTLS_E_AGAIN)
        return false;
    else if (gnutls_error_is_fatal(a))
        throw TlsError("TLS Handshake failed", a);
    return true;
}

void SecureProcess::bye()
{
    gnutls_bye(secure_impl_->session_.get(), GNUTLS_SHUT_RDWR);
}

void SecureProcess::flush()
{}


void SecureProcess::InitLibrary()
{
    if (gnutls_global_init() != GNUTLS_E_SUCCESS)
        throw PdCom::Exception("gnults_global_init() failed");
}

void SecureProcess::FinalizeLibrary()
{
    gnutls_global_deinit();
}
