/*****************************************************************************
 *
 * Copyright (C) 2021 Bjarne von Horn (vh at igh dot de)
 *
 * This file is part of the PdCom library.
 *
 * The PdCom library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * The PdCom library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the PdCom library. If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include "PeriodicSubscriptions.h"

#include "Channel.h"
#include "Subscription.h"

#include <cassert>
#include <cstring>
#include <pdcom5/Exception.h>
#include <sstream>

using PdCom::impl::MsrProto::PeriodicSubscription;
using PdCom::impl::MsrProto::PeriodicSubscriptions;


PeriodicSubscriptions::transaction PeriodicSubscriptions::add(
        PdCom::impl::Process::PendingCallbackQueue &queue,
        bool notify_pending,
        const std::shared_ptr<PeriodicSubscription> &s)
{
    const auto var_ptr = s->variable_.lock();
    if (!var_ptr)
        throw EmptyVariable();
    const auto &var     = static_cast<Channel const &>(*var_ptr);
    const auto group_id = map_.find(s->subscriber_.getTransmission());
    auto map_it         = map_.at(group_id).find(var.index_);
    if (map_it == map_.at(group_id).end()) {
        // not yet subscribed
        unsigned decimation = s->subscriber_.getTransmission().getInterval()
                        / var.sample_time_.count()
                + 0.5;
        decimation += s->subscriber_.getTransmission().getInterval() > 0.0
                and !decimation;
        unsigned blocksize = 0;

        if (decimation > 0) {
            blocksize =
                    1.0 / var.sample_time_.count() / decimation / 25.0 + 0.5;
            blocksize = std::min(var.bufsize / decimation, blocksize);
        }

        if (!blocksize)
            blocksize = 1;
        time_vector_.resize(
                std::max<size_t>(time_vector_.size(), blocksize + 2));

        map_.at(group_id).insert({var.index_, {blocksize, decimation, {{s}}}});

        if (notify_pending)
            queue.push_back({s, PdCom::Subscription::State::Pending});
        return {true, group_id, decimation, blocksize};
    }
    else {
        bool state_set = false;
        for (auto weak_other_s : map_it->second.subscriptions_) {
            if (const auto other_s = weak_other_s.lock()) {
                if (!(other_s->This_->getState()
                              == PdCom::Subscription::State::Pending
                      and !notify_pending)) {
                    queue.push_back({s, other_s->This_->getState()});
                    state_set = true;
                    break;
                }
            }
        }
        if (notify_pending and !state_set)
            queue.push_back({s, PdCom::Subscription::State::Pending});
        map_it->second.subscriptions_.insert(s);
        return {false, group_id, map_it->second.decimation_,
                map_it->second.blocksize_};
    }
}

PeriodicSubscriptions::transaction
PeriodicSubscriptions::remove(PeriodicSubscription &s)
{
    const auto var_ptr = s.variable_.lock();
    if (!var_ptr)
        throw EmptyVariable();
    const auto &var = static_cast<Channel const &>(*var_ptr);
    const auto group_id =
            map_.erase(s.subscriber_.getTransmission(), var.index_);
    return {group_id != 0, group_id, 0, 0};
}

void PeriodicSubscriptions::startNewDataRecieving(
        unsigned int index,
        PdCom::impl::MsrProto::VariableCache::Lock lock)
{
    if (!map_.valid(index))
        return;
    current_group_       = index;
    variable_cache_lock_ = std::move(lock);
}


void PeriodicSubscriptions::readData(
        const PdCom::impl::MsrProto::Channel &c,
        const char *data)
{
    if (!map_.valid(current_group_)) {
        current_group_ = -1;
        return;
    }
    const auto sub_info_it = map_.at(current_group_).find(c.index_);
    if (sub_info_it == map_.at(current_group_).end())
        return;
    if (current_blocksize_ == -1)
        current_blocksize_ = sub_info_it->second.blocksize_;
    else if (
            current_blocksize_
            != static_cast<int>(sub_info_it->second.blocksize_))
        throw ProtocolError("Blocksize mismatch");
    variable_cache_lock_[c.index_].readFromBase64(
            data, c, sub_info_it->second.blocksize_);
}


void PeriodicSubscriptions::readTimestamps(const char *data)
{
    static_assert(
            sizeof(decltype(time_vector_)::value_type) == sizeof(int64_t),
            "time_vector_ is not size compatible to time tag");
    if (current_blocksize_ == -1) {
        const auto b64len  = ::strlen(data);
        current_blocksize_ = b64len * 3 / 4 / sizeof(int64_t);
    }

    time_vector_.resize(
            std::max<size_t>(current_blocksize_ + 1, time_vector_.size()));
    if (!readFromBase64(
                reinterpret_cast<char *>(time_vector_.data()), data,
                sizeof(int64_t) * current_blocksize_)) {
        throw ProtocolError("Invalid <time> child of <data>");
    }
}

void PeriodicSubscriptions::endNewDataRecieving()
{
    // make sure to release the variable cache lock
    auto cache = std::move(variable_cache_lock_);
    // subscriptions can dissappear during <data> tag parsing, so don't
    // assume that subscriptions_ is not empty
    if (!cache or current_blocksize_ == -1)
        return;

    for (int block = 0; block < current_blocksize_; ++block) {
        for (const auto &sub_map : map_.at(current_group_)) {
            const DataDecoder *var = cache.filled_cache(sub_map.first);
            // no data recieved for given variable
            if (!var)
                continue;
            bool size_check_done = false;
            for (const auto &weak_sub : sub_map.second.subscriptions_) {
                if (const auto sub = weak_sub.lock()) {
                    const auto var_ptr = sub->variable_.lock();
                    if (!var_ptr)
                        throw EmptyVariable();
                    if (!size_check_done) {
                        if (var->size() < sub_map.second.blocksize_
                                    * var_ptr->totalSizeInBytes())

                        {
                            std::ostringstream os;
                            os << "size mismatch in variable cache, "
                                  "expected "
                               << sub_map.second.blocksize_
                                            * var_ptr->totalSizeInBytes()
                               << ", got " << var->size();
                            throw ProtocolError(os.str());
                        }
                        size_check_done = true;
                    }
                    sub->readData(
                            var->data() + block * var_ptr->totalSizeInBytes());
                    subscriber_notifier_ << sub->subscriber_;
                }
            }
        }
        subscriber_notifier_.notify(time_vector_.at(block));
    }
    subscriber_notifier_.clear();
    current_group_     = 0;
    current_blocksize_ = -1;
}

unsigned int PeriodicSubscriptions::GroupMap::find(PdCom::Transmission t)
{
    const auto it = backward_.find(t);
    if (it != backward_.end())
        return it->second;
    for (unsigned int i = 0; i < forward_.size(); ++i) {
        if (forward_[i].remove_expired()) {
            return backward_[t] = i + 1;
        }
    }
    forward_.emplace_back();
    return backward_[t] = forward_.size();
}
unsigned int PeriodicSubscriptions::GroupMap::erase(
        PdCom::Transmission t,
        unsigned int channel_idx) noexcept
{
    const auto backward_it = backward_.find(t);
    if (backward_it == backward_.end()) {
        assert(backward_it != backward_.end()
               and "Inconsistent state, backward_ and forward_ out of sync");
        return 0;
    }
    const auto map_it = forward_[backward_it->second - 1].find(channel_idx);
    if (map_it == forward_[backward_it->second - 1].end())
        return 0;
    const auto ans = map_it->second.subscriptions_.remove_expired()
            ? backward_it->second
            : 0;
    if (forward_[backward_it->second - 1].remove_expired())
        backward_.erase(backward_it);
    return ans;
}
