/*****************************************************************************
 *
 * Copyright (C) 2021 Bjarne von Horn (vh at igh dot de)
 *
 * This file is part of the PdCom library.
 *
 * The PdCom library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * The PdCom library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the PdCom library. If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#ifndef MSRPROTO_PERIODICSUBSCRIPTIONS_H
#define MSRPROTO_PERIODICSUBSCRIPTIONS_H

#include "../Process.h"
#include "../Subscription.h"
#include "DataConverter.h"

#include <chrono>
#include <memory>
#include <pdcom5/Subscriber.h>
#include <set>
#include <unordered_map>
#include <vector>

namespace PdCom { namespace impl { namespace MsrProto {
class Channel;
class PollSubscription;
class PeriodicSubscription;

template <class S = PollSubscription>
struct SubscriptionSet :
    std::set<std::weak_ptr<S>, std::owner_less<std::weak_ptr<S>>>
{
    using std::set<std::weak_ptr<S>, std::owner_less<std::weak_ptr<S>>>::set;
    void readData(const char *data, std::chrono::nanoseconds ts)
    {
        for (const auto &weak_sub : *this) {
            if (const auto sub = weak_sub.lock()) {
                sub->readData(data);
                sub->newValues(ts);
            }
        }
    }
    bool remove_expired() noexcept
    {
        for (auto it = this->begin(); it != this->end();) {
            if (it->expired())
                it = this->erase(it);
            else
                ++it;
        }
        return this->empty();
    }
};

class PeriodicSubscriptions
{
  public:
    struct transaction
    {
        bool needs_server_action;
        unsigned int group_id, decimation_, blocksize_;
    };

    struct ChannelSubscripion
    {
        unsigned int const blocksize_, decimation_;
        SubscriptionSet<PeriodicSubscription> subscriptions_;
    };

    struct ChannelSubscriptionMap :
        std::unordered_map<unsigned int /* id */, ChannelSubscripion>
    {
        using std::unordered_map<unsigned int, ChannelSubscripion>::
                unordered_map;

        bool remove_expired() noexcept
        {
            for (auto it = begin(); it != end();) {
                if (it->second.subscriptions_.remove_expired())
                    it = erase(it);
                else
                    ++it;
            }
            return empty();
        }
    };

    transaction
    add(Process::PendingCallbackQueue &q,
        bool notify_pending,
        const std::shared_ptr<PeriodicSubscription> &s);
    transaction remove(PeriodicSubscription &s);

    void startNewDataRecieving(
            unsigned int index,
            VariableCache::Lock variable_cache_lock);
    void endNewDataRecieving();

    void readData(const Channel &c, const char *data);
    void readTimestamps(const char *data);

    operator bool() const { return current_group_ and variable_cache_lock_; }

    ChannelSubscriptionMap &at(unsigned int group_id)
    {
        return map_.at(group_id);
    }

  private:
    class GroupMap
    {
        std::vector<ChannelSubscriptionMap> forward_;
        std::unordered_map<Transmission, unsigned int /* group id */> backward_;

      public:
        bool valid(unsigned int idx) const noexcept
        {
            return idx > 0 and idx <= forward_.size();
        }
        ChannelSubscriptionMap &at(unsigned int group_id)
        {
            return forward_.at(group_id - 1);
        }
        unsigned int find(Transmission t);
        unsigned int erase(Transmission t, unsigned int channel_idx) noexcept;
    } map_;
    std::vector<std::chrono::nanoseconds> time_vector_;
    VariableCache::Lock variable_cache_lock_;
    int current_blocksize_      = -1;
    unsigned int current_group_ = 0;
    impl::Subscription::SubscriberNotifier subscriber_notifier_;
};

}}}     // namespace PdCom::impl::MsrProto

#endif  // MSRPROTO_PERIODICSUBSCRIPTIONS_H
